
from __future__ import print_function

import os
import cv2
import time
import argparse
import numpy as np
import matplotlib.pyplot as plt
import pycuda.driver as cuda
import pycuda.autoinit
import tensorrt as trt
import cityscapes_onnx_trt_int8_calibrator
from cityscapes_onnx_trt_int8_calibrator import BiseNetv2EntropyCalibrator

MAX_BATCH_SIZE = 1
NET_WIDTH = 1024
NET_HEIGHT = 512

def load_onnx(model_name):
    """Read the ONNX file."""
    onnx_path = '%s.onnx' % model_name
    #onnx_path = model_name
    if not os.path.isfile(onnx_path):
        print('ERROR: file (%s) not found!  You might want to run yolo_to_onnx.py first to generate it.' % onnx_path)
        return None
    else:
        with open(onnx_path, 'rb') as f:
            return f.read()


def set_net_batch(network, batch_size):
    """Set network input batch size.

    The ONNX file might have been generated with a different batch size,
    say, 64.
    """
    if trt.__version__[0] >= '7':
        shape = list(network.get_input(0).shape)
        shape[0] = batch_size
        network.get_input(0).shape = shape
    return network


def build_engine(model_name, do_int8=True, verbose=True):
    """Build a TensorRT engine from ONNX using the older API."""
    print('Loading the ONNX file...')
    onnx_data = load_onnx(model_name)
    if onnx_data is None:
        return None

    TRT_LOGGER = trt.Logger(trt.Logger.VERBOSE) if verbose else trt.Logger()
    EXPLICIT_BATCH = [] if trt.__version__[0] < '7' else \
        [1 << (int)(trt.NetworkDefinitionCreationFlag.EXPLICIT_BATCH)]
    with trt.Builder(TRT_LOGGER) as builder, builder.create_network(*EXPLICIT_BATCH) as network, trt.OnnxParser(network, TRT_LOGGER) as parser:
        if not builder.platform_has_fast_int8:
            raise RuntimeError('INT8 not supported on this platform')
        if not parser.parse(onnx_data):
            print('ERROR: Failed to parse the ONNX file.')
            for error in range(parser.num_errors):
                print(parser.get_error(error))
            return None
        network = set_net_batch(network, MAX_BATCH_SIZE)

        print('Building an engine.  This would take a while...')
        print('(Use "--verbose" or "-v" to enable verbose logging.)')
        # if trt.__version__[0] >= '7':  # older API: build_cuda_engine()
        # new API: build_engine() with builder config
        print("tensorrt version: " + str(trt.__version__))
        builder.max_batch_size = MAX_BATCH_SIZE
        config = builder.create_builder_config()
        config.max_workspace_size = 1 << 30
        config.set_flag(trt.BuilderFlag.GPU_FALLBACK)
        config.set_flag(trt.BuilderFlag.FP16)
        profile = builder.create_optimization_profile()

        profile.set_shape(
            #'input_tensor:0',                          # input tensor name
            #(MAX_BATCH_SIZE, 3, net_h, net_w),  # min shape
            #(MAX_BATCH_SIZE, 3, net_h, net_w),  # opt shape
            #(MAX_BATCH_SIZE, 3, net_h, net_w))  # max shape
            'input_tensor:0',                          # input tensor name
            (MAX_BATCH_SIZE, NET_HEIGHT, NET_WIDTH, 3),  # min shape
            (MAX_BATCH_SIZE, NET_HEIGHT, NET_WIDTH, 3),  # opt shape
            (MAX_BATCH_SIZE, NET_HEIGHT, NET_WIDTH, 3))  # max shape
        config.add_optimization_profile(profile)
        
        if do_int8:
            print("DO INT8")
            
            config.set_flag(trt.BuilderFlag.INT8)
            config.int8_calibrator = BiseNetv2EntropyCalibrator(
                'calib_images', NET_WIDTH, NET_HEIGHT,
                'calib_%s.bin' % model_name)
            config.set_calibration_profile(profile) 
        engine = builder.build_engine(network, config) # error of IndexError: _Map_base::at

        if engine is not None:
            print('Completed creating engine.')
        return engine

# for inference
LABEL_CONTOURS = [(0, 0, 0),  # 0=road
    # 1=sidewalk, 2=building, 3=wall, 4=fence, 5=pole
    (128, 0, 0), (0, 128, 0), (128, 128, 0), (0, 0, 128), (128, 0, 128),
    # 6=traffic light, 7=traffic sign, 8=vegetation, 9=terrain, 10=sky
    (0, 128, 128), (128, 128, 128), (64, 0, 0), (192, 0, 0), (64, 128, 0),
    # 11=person, 12=rider, 13=car, 14=truck, 15=bus
    (192, 128, 0), (64, 0, 128), (192, 0, 128), (64, 128, 128), (192, 128, 128),
    # 16=train, 17=motorcycle, 18=bicycle
    (0, 64, 0), (128, 64, 0), (0, 192, 0)]

def _decode_prediction_mask(mask):
    """

    :param mask:
    :return:
    """
    mask_shape = mask.shape
    mask_color = np.zeros(shape=[mask_shape[0], mask_shape[1], 3], dtype=np.uint8)

    unique_label_ids = [v for v in np.unique(mask) if v != 0 and v != 255]

    for label_id in unique_label_ids:
        idx = np.where(mask == label_id)
        mask_color[idx] = LABEL_CONTOURS[label_id]

    return mask_color

def time_profile_trt_engine(image_file_path, trt_engine_file_path):
    TRT_LOGGER = trt.Logger(trt.Logger.ERROR)
    with open(trt_engine_file_path, 'rb') as f, trt.Runtime(TRT_LOGGER) as runtime:
        engine = runtime.deserialize_cuda_engine(f.read())

    # read input image file
    h_input = np.empty(shape=[NET_HEIGHT, NET_WIDTH, 3], dtype=np.float32)
    h_output = np.empty(shape=[NET_HEIGHT, NET_WIDTH], dtype=np.int32)

    # Alocate device memory
    d_input = cuda.mem_alloc(1 * h_input.nbytes)
    d_output = cuda.mem_alloc(1 * h_output.nbytes)
    bindings = [int(d_input), int(d_output)]

    # read images
    src_image = cv2.imread(image_file_path, cv2.IMREAD_COLOR)
    src_image = cityscapes_onnx_trt_int8_calibrator._preprocess_bisenetv2_HWCN(src_image, NET_WIDTH, NET_HEIGHT)

    loop_times = 1
    # Create a stream in which to copy inputs/outputs and run inference.
    stream = cuda.Stream()
    with engine.create_execution_context() as context:
        t_start = time.time()
        for i in range(loop_times):
            # Transfer input data to the GPU.
            cuda.memcpy_htod_async(d_input, src_image, stream)
            # Run inference.
            context.execute_async(bindings=bindings, stream_handle=stream.handle)
            # Transfer predictions back from the GPU.
            cuda.memcpy_dtoh_async(h_output, d_output, stream)
        cost_time = (time.time() - t_start) / loop_times
        # Synchronize the stream
        stream.synchronize()
        print('Inference cost: {:.5f}'.format(cost_time))
        print('Inference fps: {:.5f}'.format(1 / cost_time))

    decode_color_mask = _decode_prediction_mask(h_output)
    plt.figure('decode_color_mask')
    plt.imshow(decode_color_mask[0:, :, (2, 1, 0)])
    plt.show()

    return

def main():
    """Create a TensorRT engine for ONNX-based YOLO."""
    parser = argparse.ArgumentParser()
    parser.add_argument(
        '-v', '--verbose', action='store_true', default=True,
        help='enable verbose output (for debugging)')
    parser.add_argument(
        '-m', '--model', type=str,
        help=("bisenetv2 model"),
        default = "./checkpoint/cityscapes/bisenetv2_cityscapes_frozen")
    args = parser.parse_args()

    # set do_int8=True to create an int8 engine, do_int8=False to create a FP16 engine
    engine = build_engine(
        args.model, do_int8=True, verbose=args.verbose)
    if engine is None:
        raise SystemExit('ERROR: failed to build the TensorRT engine!')

    engine_path = '%s.trt' % args.model
    with open(engine_path, 'wb') as f:
        f.write(engine.serialize())
    print('Serialized the TensorRT engine to file: %s' % engine_path)

    time_profile_trt_engine("./src_image.jpg", "./checkpoint/cityscapes/bisenetv2_cityscapes_frozen.trt")


if __name__ == '__main__':
    main()
