import os

import numpy as np
import cv2
import pycuda.autoinit
import pycuda.driver as cuda
import tensorrt as trt
from local_utils.config_utils import parse_config_utils

CFG = parse_config_utils.cityscapes_cfg_v2

def _preprocess_bisenetv2(src_image, input_shape):
    # NCHW
    src_image = cv2.resize(src_image, dsize=target_size, interpolation=cv2.INTER_LINEAR)
    src_image = cv2.cvtColor(src_image, cv2.COLOR_BGR2RGB)
    src_image = src_image.transpose((2, 0, 1)).astype(np.float32)
    src_image = src_image.astype('float32') / 255.0
    img_mean = np.array(CFG.DATASET.MEAN_VALUE).reshape((len(CFG.DATASET.MEAN_VALUE), 1, 1))
    img_std = np.array(CFG.DATASET.STD_VALUE).reshape((len(CFG.DATASET.MEAN_VALUE),1, 1))
    src_image -= img_mean
    src_image /= img_std

    return src_image

def _preprocess_bisenetv2_HWCN(src_image, NN_input_width, NN_input_height):
    # NCHW
    src_image = src_image[:, :, (2, 1, 0)]
    src_image = cv2.resize(src_image, dsize=(NN_input_width, NN_input_height), interpolation=cv2.INTER_LINEAR)
    src_image = src_image.astype('float32') / 255.0
    img_mean = np.array(CFG.DATASET.MEAN_VALUE).reshape((1, 1, len(CFG.DATASET.MEAN_VALUE)))
    img_std = np.array(CFG.DATASET.STD_VALUE).reshape((1, 1, len(CFG.DATASET.STD_VALUE)))
    src_image -= img_mean
    src_image /= img_std

    return src_image

class BiseNetv2EntropyCalibrator(trt.IInt8EntropyCalibrator2):
    def __init__(self, img_dir, net_width, net_height, cache_file, batch_size=1):
        if not os.path.isdir(img_dir):
            raise FileNotFoundError('%s does not exist' % img_dir)

        super().__init__()

        print("BiseNetv2EntropyCalibrator info: " +
            "img_dir: " + str(img_dir) + "\n" + 
            "net_hw: " + str((net_width, net_height)) + "\n" + 
            "cache_file: " + str(cache_file) + "\n" + 
            "batch_size: " + str(batch_size) + "\n" + 
            "blob_size: " + str(3 * net_height * net_width * np.dtype('float32').itemsize * batch_size) )

        self.img_dir = img_dir
        self.net_width = net_width
        self.net_height = net_height
        self.cache_file = cache_file
        self.batch_size = batch_size
        self.blob_size = 3 * net_height * net_width * np.dtype('float32').itemsize * batch_size

        self.jpgs = [f for f in os.listdir(img_dir) if f.endswith('.jpg')]
        if len(self.jpgs) < 500:
            print('WARNING: found less than 500 images in %s!' % img_dir)
        self.current_index = 0

        # Allocate enough memory for a whole batch.
        self.device_input = cuda.mem_alloc(self.blob_size)

    def __del__(self):
        del self.device_input  # free CUDA memory

    def get_batch_size(self):
        return self.batch_size

    def get_batch(self, names):
        if self.current_index + self.batch_size > len(self.jpgs):
            return None
        current_batch = int(self.current_index / self.batch_size)

        batch = []
        for i in range(self.batch_size):
            img_path = os.path.join(
                self.img_dir, self.jpgs[self.current_index + i])
            img = cv2.imread(img_path)
            assert img is not None, 'failed to read %s' % img_path
            batch.append(_preprocess_bisenetv2_HWCN(img, self.net_width, self.net_height))
            
        batch = np.stack(batch)
        assert batch.nbytes == self.blob_size

        cuda.memcpy_htod(self.device_input, np.ascontiguousarray(batch))
        self.current_index += self.batch_size
        return [self.device_input]

    def read_calibration_cache(self):
        # If there is a cache, use it instead of calibrating again.
        # Otherwise, implicitly return None.
        if os.path.exists(self.cache_file):
            with open(self.cache_file, 'rb') as f:
                return f.read()

    def write_calibration_cache(self, cache):
        with open(self.cache_file, 'wb') as f:
            f.write(cache)
